/*
 * vifctl/iw_compat.h
 * $Rev: 1433 $
 *
 * Copyright (C) 2009 Reto Gantenbein <reto.gantenbein@gmail.com>
 * Copyright (C) 1997-2007 Jean Tourrilhes <jt@hpl.hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef IWLIB_H                     // if iwlib.h included we don't need this

/**
 * @file iw_compat.h
 * @brief Compatibility types and functions for wireless ioctl() handling
 *
 * This file provides types and functions originally defined in iwlib.h
 * from wireless-tools. Since we need compatibility to the wireless-tools
 * structures and cannot link against libiw, we have to redefine them here.
 */

#ifndef __IW_COMPAT_H
#define __IW_COMPAT_H

/* INCLUDES *******************************************************************/

#include <netinet/in.h>             // struct in_addr (for wireless.h)
#include <linux/wireless.h>         // iw_range, iw_freq, iw_param, ...
#include <net/ethernet.h>           // ether_addr
#include <string.h>                 // memcmp()


/* CONSTANTS & MACROS *********************************************************/

/** Recommended Wireless Extension version */
#define WE_VERSION	WIRELESS_EXT    // get WE version from wireless.h

/* Define error types (compatible to iwconfig.c) */
#define IWERR_ARG_NUM       -2      /**< wrong number of arguments */
#define IWERR_ARG_TYPE      -3      /**< wrong argument type */
#define IWERR_ARG_SIZE      -4      /**< wrong argument size */
#define IWERR_ARG_CONFLICT  -5      /**< conflicting arguments */
#define IWERR_SET_EXT       -6      /**< error setting iwreq */
#define IWERR_GET_EXT       -7      /**< error getting iwreq */

/* Some usefull constants */
#define KILO	1e3
#define MEGA	1e6
#define GIGA	1e9

/* Shortcuts */
typedef struct iw_statistics	iwstats;
typedef struct iw_range		iwrange;
typedef struct iw_param		iwparam;
typedef struct iw_freq		iwfreq;
typedef struct iw_quality	iwqual;
typedef struct iw_priv_args	iwprivargs;
typedef struct sockaddr		sockaddr;

/**
 * @struct wireless_config
 * Structure for storing all wireless information for each device
 */
typedef struct wireless_config
{
  char		name[IFNAMSIZ + 1]; /**< Wireless/protocol name (e.g. IEEE 802.11b) */
  int		has_nwid;
  iwparam	nwid;               /**< Network ID */
  int		has_freq;
  double	freq;               /**< Can be frequency in Hz or channel */
  int		freq_flags;
  int		has_key;
  unsigned char	key[IW_ENCODING_TOKEN_MAX]; /**< Encoding key used */
  int		key_size;           /**< Size of key in bytes */
  int		key_flags;          /**< Various key flags */
  int		has_essid;
  int		essid_on;
  char		essid[IW_ESSID_MAX_SIZE + 1];   /** ESSID (extended network) */
  int		has_mode;
  int		mode;               /**< Operation mode (e.g. master, adhoc, ...) */
}
wireless_config;

/**
 * @struct wireless_info
 * Structure for storing all wireless information for each device
 * This is pretty exhaustive...
 */
typedef struct wireless_info
{
  struct wireless_config	b;  /**< Basic information */

  int		has_sens;
  iwparam	sens;               /**< Radio sensitivity */
  int		has_nickname;
  char		nickname[IW_ESSID_MAX_SIZE + 1]; /**< Nickname */
  int		has_ap_addr;
  sockaddr	ap_addr;            /**< Access point address */
  int		has_bitrate;
  iwparam	bitrate;            /**< Bit rate in bps */
  int		has_rts;
  iwparam	rts;                /**< Ready-To-Send threshold in bytes */
  int		has_frag;
  iwparam	frag;               /**< Fragmentation threshold in bytes */
  int		has_power;
  iwparam	power;              /**< Power management parameter */
  int		has_txpower;
  iwparam	txpower;            /**< Transmit Power in dBm or mW (check flag) */
  int		has_retry;
  iwparam	retry;              /**< Retry limit or lifetime */

  /* Stats */
  iwstats	stats;
  int		has_stats;
  iwrange	range;
  int		has_range;

  /* Auth params for WPA/802.1x/802.11i */
  int		auth_key_mgmt;
  int		has_auth_key_mgmt;
  int		auth_cipher_pairwise;
  int		has_auth_cipher_pairwise;
  int		auth_cipher_group;
  int		has_auth_cipher_group;
}
wireless_info;


/* HELPER SUBROUTINES *********************************************************/
int
iw_get_param_state(struct iw_param * par);


/* ADDRESS SUBROUTINES ********************************************************/

/**
 * Compare two ethernet addresses.
 */
static inline int
iw_ether_cmp(const struct ether_addr* eth1, const struct ether_addr* eth2)
{
    return memcmp(eth1, eth2, sizeof(*eth1));
}

void
iw_ether_ntop(const struct ether_addr *eth, char *buf);

char *
iw_sawap_ntop(const struct sockaddr *sap, char *buf);


/* FREQUENCY SUBROUTINES ******************************************************/

void
iw_float2freq(double in, iwfreq *out);

double
iw_freq2float(const iwfreq *in);


/* TXPOWER SUBROUTINES ********************************************************/

int16_t
dbm2mw(int16_t in);

int16_t
iw_mwatt2dbm(int16_t in);

#endif  // __IW_COMPAT_H
#endif  // IWLIB_H
