/*
 * vifctl/vif_access.c
 * $Rev: 1433 $
 *
 * Copyright (C) 2009-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file vif_access.c
 * @brief Functions for handling virtual interface access
 */

/* INCLUDES *******************************************************************/

#include <sys/ioctl.h>                  // ioctl()
#include <sys/stat.h>
#include <linux/if_tun.h>               // IFF_TUN, TUNSETIFF, ...
#include <errno.h>                      // errno, EPERM, EINVAL, ...
#include <fcntl.h>                      // open()
#include <stdio.h>                      // fprintf(), perror()
#include <stdlib.h>                     // calloc(), free()
#include <string.h>                     // strncpy(), memset()
#include <unistd.h>                     // geteuid(), getegid(), close()
#include "iw_compat.h"                  // wireless_info, iw_float2freq(), ...
#include "vif.h"


/* CONFIG ACCESS FUNCTIONS ****************************************************/

/* vif_config.c */
extern int vif_config_load(void);
extern int vif_config_create(void);
extern int vif_config_list(int, char[MAX_NUM_VIF][IFNAMSIZ + 1]);
extern vif_info *vif_config_vif_info(int, const char *);
extern int vif_config_add(int, const char *);
extern int vif_config_delete(int, const char *);
extern void vif_config_print(int);

/* vif_notify.c */
extern int vif_notify_ioctl(int, const char *);


/* LOCAL VARIABLES ************************************************************/

/** Virtual interface configuration list identifier */
int vif_list_id = -1;


/* OUTPUT *********************************************************************/

/** Define empty function with printf() signature */
int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Debug output function (ignored by default)
 *  can be changed with vif_set_debug_function() */
int (*vif_debug)(const char *format, ...) = &ignore_debug;

/** Error output function (default to stdout)
 *  can be changed with vif_set_error_function() */
int (*vif_error)(const char *format, ...) = &printf;


/* VIF DEVICE FUNCTIONS *******************************************************/

/**
 * Creates new TUN interfaces. Implementation adapted from tunctl.c
 * (uml_utilities);
 * Copyright: 2002, Jeff Dike
 */
int
tun_create(const char* dev_name)
{
    struct ifreq ifr;
    int tap_fd = -1;
    int rtrn = -1;
    char *file = "/dev/net/tun";
    uid_t owner = geteuid();
    gid_t group = getegid();

    if ((tap_fd = open(file, O_RDWR)) < 0)
    {
        vif_error("Failed to open %s (%s)\n", file, strerror(errno));
        vif_error("Is 'tun' kernel module loaded?\n");
        return -1;
    }

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    strncpy(ifr.ifr_name, dev_name, sizeof(ifr.ifr_name) - 1);

    if ((rtrn = ioctl(tap_fd, TUNSETIFF, (void *) &ifr)) < 0)
    {
        switch (errno)
        {
            case EPERM:
                vif_error("No permission to create TUN interface\n");
                break;
            case EINVAL:
                vif_error("Invalid device name %s\n", dev_name);
                break;
            default:
                vif_error("tun_create(): TUNSETIFF");
                break;
        }
        goto finish;
    }
    if ((rtrn = ioctl(tap_fd, TUNSETOWNER, owner)) < 0)
    {
        vif_error("tun_create(): TUNSETOWNER");
        goto finish;
    }
    if ((rtrn = ioctl(tap_fd, TUNSETGROUP, group)) < 0)
    {
        vif_error("tun_create(): TUNSETGROUP");
        goto finish;
    }
    if ((rtrn = ioctl(tap_fd, TUNSETPERSIST, 1)) < 0)
    {
        vif_error("tun_create(): enabling TUNSETPERSIST");
        goto finish;
    }

finish:
    close(tap_fd);
    return rtrn;
}

/** Remove TUN interface */
int
tun_remove(const char* dev_name)
{
    struct ifreq ifr;
    int tap_fd = -1;
    int rtrn = -1;
    char *file = "/dev/net/tun";

    if ((tap_fd = open(file, O_RDWR)) < 0)
        return -1;

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    strncpy(ifr.ifr_name, dev_name, sizeof(ifr.ifr_name) - 1);

    if ((rtrn = ioctl(tap_fd, TUNSETIFF, (void *) &ifr)) == 0)
    {
        if ((rtrn = ioctl(tap_fd, TUNSETPERSIST, 0)) < 0)
        {
            vif_error("Warning: Couldn't remove %s (%s)",
                dev_name, strerror(errno));
            vif_error("Notice: Manually delete device with `tunctl -d %s`\n",
                dev_name);
        }
    }

    close(tap_fd);
    return rtrn;
}


/* VIF HELPER FUNCTIONS *******************************************************/

/** Return wireless_info of virtual interface */
wireless_info *
get_wireless_info(const char *dev_name)
{
    vif_info *vif_entry = vif_config_vif_info(vif_list_id, dev_name);
    //~ printf("get_wireless_info(): dev_name = %s, vif_name = %s\n", dev_name, vif_entry->vif_name);
    //~ printf("                     txpower = %i\n", vif_entry->vif_wlaninfo.txpower.value);

    return &vif_entry->vif_wlaninfo;
}


/* IOCTL SUB-FUNCTIONS ))******************************************************/

int
vif_siocsifhwaddr(const char *dev_name)
{
    int skfd = -1;
    struct ifreq ifr;
    vif_info *vif_entry = NULL;

    if ((skfd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        vif_error("Cannot open socket (%s)", strerror(errno));
        return -1;
    }
    strncpy(ifr.ifr_name, dev_name, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFHWADDR, &ifr) != 0)
    {
        vif_error("Failed to get MAC address of %s (%s)\n",
            dev_name, strerror(errno));
        close(skfd);
        return -1;
    }
    close(skfd);

    if ((vif_entry = vif_config_vif_info(vif_list_id, dev_name)) == NULL)
        return -1;
    memmove(vif_entry->vif_mac, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

    return 0;
}

int
vif_siocgiwname(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(request->u.name, vif_config->b.name, IFNAMSIZ);

    vif_debug("SIGCGIWNAME : %s\n", request->u.name);
    return 0;
}

int
vif_siocgiwnid(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    if (vif_config->b.has_nwid == 0)
    {
        vif_debug("SIOCGIWNWID : off\n");
        return IWERR_GET_EXT;
    }
    memmove(&(request->u.nwid), &vif_config->b.nwid, sizeof(struct iw_param));
    vif_config->b.has_nwid = iw_get_param_state(&(vif_config->b.nwid));

    vif_debug("SIOCGIWNWID : %c\n", request->u.nwid.value);
    return 0;
}

/** Called for SIOCSIWFREQ (argument can be channel or frequency) */
int
vif_siocsiwfreq(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    vif_config->b.freq_flags = request->u.freq.flags;
    if (vif_config->b.freq_flags == IW_FREQ_AUTO)   // reset frequency
    {
        vif_config->b.has_freq = 0;
        vif_config->b.freq = channel_frequency[0] * MEGA;
    } else {                                // set new frequency
        vif_config->b.has_freq = 1;
        float freq = iw_freq2float(&(request->u.freq));

        // if freq < 1000 we assume it's a channel
        if (freq < MEGA)
        {
            int channel = (int) freq;
            if (channel > vif_config->range.num_channels - 1)
                return IWERR_ARG_SIZE;
            vif_config->b.freq = channel_frequency[channel] * MEGA;
        } else {
            vif_config->b.freq = freq;
        }
    }

    if (vif_config->b.freq > 1000)
    {
        vif_debug("SIOCSIWFREQ : %.3e GHz\n", vif_config->b.freq / GIGA);
    } else {
        vif_debug("SIOCSIWFREQ : Channel %.0f\n", vif_config->b.freq);
    }
    return 0;
}

int
vif_siocgiwfreq(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    iw_float2freq(vif_config->b.freq, &(request->u.freq));

    if (vif_config->b.freq > 1000)
    {
        vif_debug("SIOCGIWFREQ : %.3e GHz\n", vif_config->b.freq / GIGA);
    } else {
        vif_debug("SIOCGIWFREQ : Channel %.0f\n", vif_config->b.freq);
    }
    return 0;
}

int
vif_siocsiwmode(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    vif_config->b.has_mode = 1;
    vif_config->b.mode = request->u.mode;

    vif_debug("SIOCSIWMODE : %i\n", vif_config->b.mode);
    return 0;
}

int
vif_siocgiwmode(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    if (vif_config->b.has_mode == 1)
    {
        request->u.mode = vif_config->b.mode;
    } else {
        request->u.mode = IW_MODE_AUTO;
    }

    vif_debug("SIOCGIWMODE : %i\n", request->u.mode);
    return 0;
}

int
vif_siocsiwsens(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    memmove(&(vif_config->sens), &(request->u.sens), sizeof(struct iw_param));
    vif_config->has_sens = iw_get_param_state(&(vif_config->sens));

    if (vif_config->sens.value < 0)
        vif_debug("SIOCSIWSENS : %i dB\n", vif_config->sens.value);
    else
        vif_debug("SIOCSIWSENS : %i\n", vif_config->sens.value);
    return 0;
}

int
vif_siocgiwsens(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.sens), &(vif_config->sens), sizeof(struct iw_param));

    if (vif_config->sens.value < 0)
        vif_debug("SIOCGIWSENS : %i dB\n", vif_config->sens.value);
    else
        vif_debug("SIOCGIWSENS : %i\n", vif_config->sens.value);
    return 0;
}

int
vif_siocgiwrange(struct iwreq *request)
{
    //~ printf("vif_siocgiwrange()\n");
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(request->u.data.pointer, &(vif_config->range), sizeof(struct iw_range));
    request->u.data.length = sizeof(struct iw_range);

    vif_debug("SIOCGIWRANGE : we_version_compiled = %i\n",
        ((struct iw_range *) request->u.data.pointer)->we_version_compiled);
    return 0;
}

int
vif_siocgiwpriv(struct iwreq *request)
{
    int i = -1;
    struct iw_priv_args *priv = request->u.data.pointer;

    if (VIF_IW_NUM_IOCTL_PRIV > request->u.data.length) // buffer too small
    {
        // send number of private ioctls to adjust buffer
        request->u.data.length = VIF_IW_NUM_IOCTL_PRIV;
        return IWERR_ARG_SIZE;
    }

    for (i = 0; i < VIF_IW_NUM_IOCTL_PRIV; i++)     // copy private ioctls
        memmove(&priv[i], &vif_private_args[i], sizeof(struct iw_priv_args));
    request->u.data.length = VIF_IW_NUM_IOCTL_PRIV;

    vif_debug("SIOCGIWPRIV : %i ioctls:\n", VIF_IW_NUM_IOCTL_PRIV);
    for (i = 0; i < VIF_IW_NUM_IOCTL_PRIV; i++)
    {
        vif_debug("   (%p) cmd: %.4X\n", &(priv[i].cmd), priv[i].cmd);
        vif_debug("   (%p) name: %s\n", priv[i].name, priv[i].name);
    }
    return 0;
}

int
vif_siocgiwap(struct iwreq *request)
{
    char buffer[sizeof(struct sockaddr)];
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.ap_addr.sa_data), &(vif_config->ap_addr.sa_data),
        ETH_ALEN);

    vif_debug("SIGCGIWAP : %s\n", iw_sawap_ntop(&(request->u.ap_addr), buffer));
    return 0;
}

int
vif_siocsiwessid(struct iwreq *request)
{
    //~ printf("vif_siocsiwessid()\n");
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    memset(vif_config->b.essid, 0, IW_ESSID_MAX_SIZE + 1);
    strncpy(vif_config->b.essid, request->u.essid.pointer, request->u.essid.length);
    vif_config->b.essid_on = request->u.essid.flags;
    vif_config->b.has_essid = request->u.essid.flags;

    vif_debug("SIOCSIWESSID : \"%s\"\n", (char *) vif_config->b.essid);
    return 0;
}

int
vif_siocgiwessid(struct iwreq *request)
{
    //~ printf("vif_siocgiwessid()\n");
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    strncpy(request->u.essid.pointer, vif_config->b.essid, IW_ESSID_MAX_SIZE + 1);
    request->u.essid.length = strlen(vif_config->b.essid) + 1;  // include trailing '\0'
    request->u.essid.flags = vif_config->b.essid_on;

    vif_debug("SIOCGIWESSID : \"%s\", u.data.length = %i\n",
        request->u.essid.pointer, request->u.essid.length);
    return 0;
}

int
vif_siocgiwnickn(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    // nickname uses same length as ESSID (see wireless.h)
    strncpy(request->u.data.pointer, vif_config->nickname, IW_ESSID_MAX_SIZE + 1);
    request->u.data.length = strlen(vif_config->nickname) + 1;  // include trailing '\0'

    vif_debug("SIOCGIWNICKN : \"%s\"\n", request->u.data.pointer);
    return 0;
}

int
vif_siocsiwrate(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    memmove(&(vif_config->bitrate), &(request->u.bitrate),
        sizeof(struct iw_param));
    vif_config->has_bitrate = iw_get_param_state(&(vif_config->bitrate));

    vif_debug("SIOCSIWRATE : %.1f Mbit/s\n", vif_config->bitrate.value / MEGA);
    return 0;
}

int
vif_siocgiwrate(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.bitrate), &(vif_config->bitrate),
        sizeof(struct iw_param));

    vif_debug("SIOCGIWRATE : %.1f Mbit/s\n", request->u.bitrate.value / MEGA);
    return 0;
}

int
vif_siocsiwrts(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    if (request->u.rts.value > VIF_MAX_RTS)
        return IWERR_ARG_SIZE;
    if (request->u.rts.value < 0)
        return IWERR_ARG_SIZE;

    memmove(&(vif_config->rts), &(request->u.rts), sizeof(struct iw_param));
    vif_config->has_rts = iw_get_param_state(&(vif_config->rts));

    if (vif_config->has_rts == 1)
        vif_debug("SIOCSIWRTS : %i Bytes\n", request->u.rts.value);
    else
        vif_debug("SIOCSIWRTS : off\n");
    return 0;
}

int
vif_siocgiwrts(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.rts), &(vif_config->rts), sizeof(struct iw_param));

    if (vif_config->has_rts == 1)
    {
        vif_debug("SIOCGIWRTS : %i Bytes\n", request->u.rts.value);
    } else {
        vif_debug("SIOCGIWRTS : off\n");
    }
    return 0;
}

int
vif_siocsiwfrag(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    //~ printf("vif_siocsiwfrag():\n");
    //~ printf("                    frag.value = %i\n", request->u.frag.value);
    //~ printf("                    frag.fixed = %i\n", request->u.frag.fixed);
    //~ printf("                    frag.disabled = %i\n", request->u.frag.disabled);
    //~ printf("                    frag.flags = %#.4x\n", request->u.frag.flags);

    // check boundaries
    // (value -1 for disabling fragmentation)
    if (request->u.frag.value < 256 && request->u.frag.value > -1)
        return IWERR_ARG_SIZE;
    if (request->u.frag.value > 2346)
        return IWERR_ARG_SIZE;

    memmove(&(vif_config->frag), &(request->u.frag), sizeof(struct iw_param));
    vif_config->has_frag = iw_get_param_state(&(vif_config->frag));

    if (vif_config->has_frag == 1)
    {
        vif_debug("SIOCSIWFRAG : %i Bytes\n", request->u.frag.value);
    } else {
        vif_debug("SIOCSIWFRAG : off\n");
    }
    return 0;
}

int
vif_siocgiwfrag(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.frag), &(vif_config->frag), sizeof(struct iw_param));

    if (vif_config->has_frag == 1)
    {
        vif_debug("SIOCGIWFRAG : %i Bytes\n", request->u.frag.value);
    } else {
        vif_debug("SIOCGIWFRAG : off\n");
    }
    return 0;
}

int
vif_siocsiwtxpow(struct iwreq *request)
{
    //~ printf("vif_siocsiwtxpow(): ");
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    // maximum 20dBm allowed
    if ((request->u.txpower.flags == IW_TXPOW_DBM) && (request->u.txpower.value > 20))
       return IWERR_ARG_SIZE;
    // maximum 100mW allowed
    if ((request->u.txpower.flags == IW_TXPOW_MWATT) && (request->u.txpower.value > 100))
       return IWERR_ARG_SIZE;

    memmove(&(vif_config->txpower), &(request->u.txpower),
        sizeof(struct iw_param));

    // check for 'iwconfig txpower auto'
    if (vif_config->txpower.value == -1)
        vif_config->txpower.value = VIF_DEFAULT_TXPWR;

    vif_config->has_txpower = iw_get_param_state(&(vif_config->txpower));

    if (vif_config->txpower.flags == IW_TXPOW_DBM)
        vif_debug("SIOCSIWTXPOW : %d dBm\n", vif_config->txpower.value);
    else
        vif_debug("SIOCSIWTXPOW : %d mW\n", vif_config->txpower.value);
    return 0;
}

int
vif_siocgiwtxpow(struct iwreq *request)
{
    //~ printf("vif_siocgiwtxpow()\n");
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.txpower), &(vif_config->txpower),
        sizeof(struct iw_param));

    if (vif_config->txpower.flags == IW_TXPOW_DBM)
        vif_debug("SIOCGIWTXPOW : %d dBm\n", vif_config->txpower.value);
    else
        vif_debug("SIOCGIWTXPOW : %d mW\n", vif_config->txpower.value);
    return 0;
}

int
vif_siocsiwretry(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);

    if (request->u.retry.value > VIF_MAX_RETRY)
        return IWERR_ARG_SIZE;
    if (request->u.retry.value < 0)
        return IWERR_ARG_SIZE;

    memmove(&(vif_config->retry), &(request->u.retry),
            sizeof(struct iw_param));
    if (vif_config->retry.value == 0)
        vif_config->retry.disabled = 1;
    else
        vif_config->retry.disabled = 0;
    vif_config->has_retry = iw_get_param_state(&(vif_config->retry));

    vif_debug("SIOCSIWRETRY : %.1d\n", request->u.retry.value);
    return 0;
}

int
vif_siocgiwretry(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);

    if (vif_config == NULL)
        return IWERR_GET_EXT;

    memmove(&(request->u.retry), &(vif_config->retry), sizeof(struct iw_param));

    vif_debug("SIOCGIWRETRY : %.1d\n", request->u.retry.value);
    return 0;
}

int
vif_siocsiwencode(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_SET_EXT;

    vif_config->b.key_flags = request->u.encoding.flags;
    vif_config->b.key_size = request->u.encoding.length;
    if (request->u.encoding.flags != IW_ENCODE_DISABLED)
    {
        memmove(vif_config->b.key, request->u.encoding.pointer,
            IW_ENCODING_TOKEN_MAX);
        vif_config->b.has_key = 1;

        vif_debug("SIOCSIWENCODE : %s, size = %i\n", vif_config->b.key,
            vif_config->b.key_size);
    } else {
        memset(vif_config->b.key, 0, IW_ENCODING_TOKEN_MAX);
        vif_config->b.has_key = 0;

        vif_debug("SIOCSIWENCODE : off\n");
    }
    return 0;
}

int
vif_siocgiwencode(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);
    if (vif_config == NULL)
        return IWERR_GET_EXT;

    if (vif_config->b.has_key == 1)
    {
        memmove(request->u.encoding.pointer, vif_config->b.key,
            IW_ENCODING_TOKEN_MAX);
        request->u.encoding.length = vif_config->b.key_size;
        request->u.encoding.flags = vif_config->b.key_flags;

        vif_debug("SIOCGIWENCODE : %s, size = %i\n",
            request->u.encoding.pointer, request->u.encoding.length);
    } else {
        request->u.encoding.flags = IW_ENCODE_DISABLED;
        request->u.encoding.length = 0;

        vif_debug("SIOCGIWENCODE : off\n");
    }
    return 0;
}

int
vif_siocgiwpower(struct iwreq *request)
{
    struct wireless_info *vif_config =
        get_wireless_info(request->ifr_ifrn.ifrn_name);

    memmove(&(request->u.power), &(vif_config->power), sizeof(struct iw_param));
    if (vif_config->has_power == 1)
    {
        vif_debug("SIGCGIWPOWER : enabled\n");
    } else {
        vif_debug("SIGCGIWPOWER : disabled'\n");
    }

    return 0;
}


/* VIF PUBLIC FUNCTIONS (see header) ******************************************/

void
vif_set_debug_function(int (*debug)(const char *format, ...))
{
    vif_debug = debug;
}

void
vif_set_error_function(int (*error)(const char *format, ...))
{
    vif_error = error;
}

ssize_t
vif_init()
{
    ssize_t num_vif = 0;
    vif_list_id = vif_config_load();

    if (vif_list_id < 0)
        return 0;
    else
        num_vif = vif_config_list(vif_list_id, NULL);

    vif_debug("vif_init(): num_vif = %i\n", num_vif);
    if (num_vif < 0)
        return 0;
    else
        return num_vif;
}

int
vif_create(const char *vif_name)
{
    int i = 0;
    int num_vif = 0;
    char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1];

    if (vif_list_id < 0)            // initialize shared memory for R/W access
        if ((vif_list_id = vif_config_create()) < 0)
            return -1;

    memset(vif_names, 0, MAX_NUM_VIF);
    num_vif = vif_config_list(vif_list_id, vif_names);

    if (num_vif == MAX_NUM_VIF)
    {
        vif_error("Failed to create interface %s. Only %i virtual interfaces supported\n",
            vif_name, MAX_NUM_VIF);
        return -1;
    }

    if (tun_create(vif_name) != 0)          // create network interface
        return -1;

    for (i = 0; i < num_vif; i++)
    {
        if (strncmp(vif_name, vif_names[i], IFNAMSIZ) == 0)
        {
            vif_error("Device %s already exists\n", vif_name);
            return -1;
        }
    }
    if (vif_config_add(vif_list_id, vif_name) < 0)
        return -1;

    vif_siocsifhwaddr(vif_name);            // set MAC address in vif_info

    return 0;
}

int
vif_delete(const char *dev_name)
{
    int rtrn = 0;

    rtrn = vif_config_delete(vif_list_id, dev_name);
    if (rtrn == 0)
        rtrn = tun_remove(dev_name);

    return rtrn;
}

int
vif_ioctl(int request, struct iwreq *pwrq)
{
    int i = 0, rtrn = -1;
    vif_info *vif_entry = NULL;

    if (vif_list_id < 0)
        vif_list_id = vif_config_load();

    vif_entry = vif_config_vif_info(vif_list_id, pwrq->ifr_name);
    if (vif_entry == NULL)                      // invalid device
    {
        vif_debug("No such device (%s)\n", pwrq->ifr_name);
        if (IW_IS_GET(request))
            return IWERR_GET_EXT;
        else
            return IWERR_SET_EXT;
    }

    for (i = 0; i < VIF_NUM_IOCTL; i++)
    {
        if (ioctl_handler_list[i].ioctl_num == request)
        {
            // execute corresponding vif_sioc*() function
            rtrn = (*ioctl_handler_list[i].iw_handler)(pwrq);
            if (rtrn < 0)
                return rtrn;
            if (IW_IS_SET(request))
                vif_notify_ioctl(request, pwrq->ifr_name);  // send notification
            return 0;                                       // now we're done!
        }
    }

    // only reach this if reqested ioctl is not in our list
    vif_debug("vif_ioctl(): requested ioctl not found\n");
    if (IW_IS_GET(request))
        return IWERR_GET_EXT;
    else
        return IWERR_SET_EXT;
}

int
vif_get_names(char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1])
{
    if (vif_list_id < 0)
        vif_list_id = vif_config_load();        // try to load a config

    if (vif_list_id < 0)
    {
        vif_error("Couldn't load virtual interface(s) configuration\n");
        return -1;
    } else {
        return vif_config_list(vif_list_id, vif_names);
    }
}

void
vif_print(void)
{
    if (vif_list_id < 0)
        vif_list_id = vif_config_load();

    vif_config_print(vif_list_id);
}

int
vif_get_mac(unsigned char *mac, const char *dev_name)
{
    vif_info *vif_entry = vif_config_vif_info(vif_list_id, dev_name);
    if (vif_entry == NULL)
        return -1;
    else
        memmove(mac, vif_entry->vif_mac, ETH_ALEN);

    return 0;
}

double
vif_get_frequency(const char *dev_name)
{
    //~ printf("vif_get_frequency()\n");
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
        return wlaninfo->b.freq;
    else
        return 0;
}

uint16_t
vif_get_channel(const char *dev_name)
{
    //~ printf("vif_get_channel()\n");
    int i = 0;
    uint16_t channel = 0;
    int freq = vif_get_frequency(dev_name) / MEGA;
    if (freq > 0)
    {
        wireless_info *wlaninfo = get_wireless_info(dev_name);
        if (wlaninfo != NULL)
        {
            for (i = 0; i < VIF_NUM_CHANNEL; i++)
            {
                int channel_freq = iw_freq2float(&(wlaninfo->range.freq[i])) / MEGA;
                if (freq == channel_freq)
                {
                    channel = i;
                    break;
                }
            }
        }
    }
    return channel;
}

int16_t
vif_get_sensitivity(const char *dev_name)
{
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
        // if sensitivity is not negative we don't understand the unit;
        // from manpage: "Positive values are assumed to be the raw value used
        // by the hardware or a percentage"
        if (wlaninfo->sens.value < 0)
        {
            // bypass bug in iw_dbm2mw() which only allows positive values
            int16_t dbm_sens = -1 * wlaninfo->sens.value;
            int16_t mw_sens = -1 * iw_dbm2mw(dbm_sens);
            return mw_sens;
        } else
            return wlaninfo->sens.value;
    else
        return 0;
}

int16_t
vif_get_rts(const char *dev_name)
{
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
        // if we have RTS/CTS threshold disabled, send maximal possible value
        if (wlaninfo->rts.disabled == 1)
            return VIF_MAX_RTS;
        else
            return wlaninfo->rts.value;
    else
        return 0;
}

int16_t
vif_get_frag(const char *dev_name)
{
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
        return wlaninfo->frag.value;
    else
        return 0;
}

int16_t
vif_get_txpower(const char *dev_name)
{
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
    {
        struct iw_param *txpwr = &(wlaninfo->txpower);
        if (txpwr->flags)
            // txpower flag is IW_TXPOW_MWATT(1), simply copy
            return txpwr->value;
        else
            // txpower flag is IW_TXPOW_DBM(0) first convert
            return iw_dbm2mw(txpwr->value);
    } else
        return 0;
}

int16_t
vif_get_retry(const char *dev_name)
{
    wireless_info *wlaninfo = get_wireless_info(dev_name);
    if (wlaninfo != NULL)
        return wlaninfo->retry.value;
    else
        return 0;
}
