/*
 * vifctl/vifctl.c
 * $Rev: 1743 $
 *
 * Copyright (C) 2008-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file vifctl.c
 * @brief Vifctl - Virtual Interface Control
 *
 * Usage:  vifctl [-cdhlr] DEVICENAME
 *         Manage virtual wireless interfaces
 * vifctl -c VIFNAME Create virtual wireless interface
 *        -r VIFNAME Remove virtual wireless interface
 *        -l         List virtual wireless interfaces
 *        -d         Enable verbose output
 *        -h         Print this help
 *
 *
 * Manage virtual wireless interfaces. These consist of a real Ethernet device
 * via TUN/TAP driver and a seperate wireless configuration. This one is
 * consistent with an unconfigured Atheros WLAN device.
 *
 * To set and get the wireless configuration you have to use a patched version
 * of the wireless-tools package.
 *
 * The devices will be automatically created at program start and deleted again
 * at program exit.
 */

#include <errno.h>                  // EINVAL
#include <stdarg.h>                 // va_list, va_start(), va_end(), vfprintf(), ...
#include <stdlib.h>                 // exit()
#include <stdio.h>                  // printf(), fopen(), fclose(), ...
#include <string.h>                 // strrchr(), strlen()
#include <syslog.h>                 // openlog(), vsyslog(), LOG_PID, ...
#include <unistd.h>                 // getopt(), fork()
#include "vif.h"                    // vif_create(), vif_handle_iwreq(), ...
#include "vifctl_print.h"           // vifctl_debug(), vifctl_debug_args(), ...


char *exec_name = NULL;                     /**< Program name */
char vifctl_version[] = "1.0\0";   /**< Program version */
int vifctl_dbg = 0;     /**< Debug mode */
int create_vif = 0;     /**< Create virtual interface */
int delete_vif = 0;     /**< Delete virtual interface */
int list_vif = 0;       /**< List virtual interfaces */
int notify_id = -1;     /**< Virtual interface notification ID */
const char *arg_name = NULL;    /**< Device name argument pointer */

extern int (*vifctl_debug)(const char* format, ...);
extern void (*vifctl_debug_args)(const char* format, va_list args);


/* FUNCTIONS ******************************************************************/

/** Print program usage to stdout */
void
vifctl_usage(int rtrn)
{
    printf("Usage:\t%s [-cdhlr] VIFNAME\n", exec_name);
    printf("\tManage virtual wireless interfaces - Vifctl v%s\n", PACKAGE_VERSION);
    printf("%s\t-c VIFNAME\tCreate virtual wireless interface\n", exec_name);
    printf("\t-r VIFNAME\tRemove virtual wireless interface\n");
    printf("\t-l\t\tList virtual wireless interfaces\n");
    printf("\t-d\t\tEnable verbose output\n");
    printf("\t-h\t\tPrint this help\n");
    exit(rtrn);
}


/** Empty function to match printf signature */
int
ignore_debug(const char* text, ...)
{
    return 0;
}


/** Print debug message with variable argument list */
int
print_debug(const char *format, ...)
{
    va_list arg;

    va_start(arg, format);
    vifctl_debug_args(format, arg);
    va_end(arg);

    return 0;
}


/** Empty function to match print_debug_args signature */
void
ignore_debug_args(const char *format, va_list arg)
{
}


/** Print debug message with given argument list */
void
print_debug_args(const char *format, va_list arg)
{
    FILE *stdout = fopen("/dev/stdout", "a");
//    fprintf(stdout, "DEBUG: ");
    vfprintf(stdout, format, arg);
    fclose(stdout);
}


int
vifctl_error(const char *format, ...)
{
    int chars_printed = 0;
    va_list arg;

    FILE *stderr = fopen("/dev/stderr", "a");
    if (stderr == NULL)
        return -1;
    chars_printed = fprintf(stderr, "Error: ");
    va_start(arg, format);
    chars_printed += vfprintf(stderr, format, arg);
    va_end(arg);
    fclose(stderr);

    return chars_printed;
}


void
vifctl_log(const char *format, ...)
{
    va_list arg;

    // send to syslog
    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    closelog();
    va_end(arg);

    // also print to stdout
    FILE *stdout = fopen("/dev/stdout", "a");
    if (stdout == NULL)
        return;
    va_start(arg, format);
    vfprintf(stdout, format, arg);
    va_end(arg);
    fclose(stdout);
}


/** Check parameter validity and return interface names and number */
ssize_t
vifctl_parse_args(int argc, char *argv[])
{
    int opt = 0;
    int num_args = argc - 1;
    ssize_t list_size = 0;

    // IMPORTANT: don't use vifctl_log or vifctl_error before here!
    vifctl_debug = ignore_debug;
    vifctl_debug_args = ignore_debug_args;

    // check for missing arguments
    if (num_args == 0)
    {
        vifctl_error("Missing argument\n");
        vifctl_usage(EXIT_FAILURE);
    }

    // check for debugging option
    while ((opt = getopt(argc, argv, "dhc:r:l")) != -1)
    {
        switch (opt)
        {
            case 'd':
                vifctl_dbg = 1;
                break;
            case 'h':
                vifctl_usage(EXIT_SUCCESS);
                break;
            case 'c':
                if (delete_vif == 1)
                {
                    vifctl_error("create+remove? Make up your mind!\n");
                    exit(EINVAL);
                }
                create_vif = 1;
                arg_name = optarg;
                break;
            case 'r':
                if (create_vif == 1)
                {
                    vifctl_error("create+remove? Make up your mind!\n");
                    exit(EINVAL);
                }
                delete_vif = 1;
                arg_name = optarg;
                break;
            case 'l':
                list_vif = 1;
                break;
        }
    }

    // set correct output/debug functions (also for libraries)
    if (vifctl_dbg == 1)
    {
        vifctl_debug = print_debug;
        vif_set_debug_function(vifctl_debug);
        vifctl_debug_args = print_debug_args;
    }
    vif_set_error_function(vifctl_error);

    // check for argument length
    if (arg_name != NULL)
    {
        if (strlen(arg_name) > IFNAMSIZ)
        {
            vifctl_error("Interface name too long\n");
            exit(EXIT_FAILURE);
        }
        if (strlen(arg_name) < 2)
        {
            vifctl_error("Interface name '%s' too short\n", arg_name);
            exit(EXIT_FAILURE);
        }
    }

    list_size = vif_init();         // access libvif
    if ((list_size == MAX_NUM_VIF) && (create_vif == 1))
    {
        vifctl_error("Already have maximum of %i interfaces\n", MAX_NUM_VIF);
        exit(EXIT_FAILURE);
    }
    if ((list_size == 0) && (delete_vif == 1))
    {
        vifctl_error("No virtual wireless interfaces found\n");
        exit(EXIT_FAILURE);
    }

    if ((arg_name == NULL) && (list_vif == 0))
    {
        vifctl_error("Missing Argument, need at least -c, -r or -l\n");
        vifctl_usage(EXIT_FAILURE);
    }

    return list_size;
}


/** MAIN */
int
main(int argc, char *argv[])
{
    int i = 0;
    int rtrn = 0;
    char name_list[MAX_NUM_VIF][IFNAMSIZ + 1];
    int list_size = 0;

    // extract executable name
    if (strrchr(argv[0], '/') == NULL)
        exec_name = argv[0];
    else
        exec_name = strrchr(argv[0], '/') + 1;

    // check arguments and examine number of virtual interfaces
    list_size = vifctl_parse_args(argc, argv);

    if (create_vif == 1)
    {
        rtrn = vif_create(arg_name);
        if (rtrn == 0)
            vifctl_log("Virtual wireless interface %s created\n", arg_name);
    }
    else if (delete_vif == 1)
    {
        rtrn = vif_delete(arg_name);
        if (rtrn == 0)
            vifctl_log("Interface %s deleted\n", arg_name);
        else
            vifctl_error("Failed to delete %s\n", arg_name);
    }
    if (list_vif == 1)
    {
        if (list_size == 0)
        {
            printf("No virtual wireless interfaces found\n");
            rtrn = EXIT_FAILURE;
        } else {
            rtrn = vif_get_names(name_list);
            if (rtrn < 1)
            {
                vifctl_error("main(): vif_get_names() failed\n");
                rtrn = EXIT_FAILURE;
            } else {
                printf("Virtual wireless interfaces:\t");
                for (i = 0; i < rtrn; i++)
                    printf("%s ", name_list[i]);
                printf("\n");
            }
        }
    }

    return rtrn;
}
